/*
 *    ImageServlet
 *  Copyright (C) 2008 Mark Elam
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 *  For more information about the GNU Public license see:
 *  http://www.gnu.org/copyleft/gpl.html
 *
 *  ImageServlet is at:
 *  http://www.subtrack.org.uk
 *
 */
package uk.org.subtrack.imaging;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import javax.imageio.ImageIO;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xmlbeans.XmlException;
import uk.org.subtrack.imaging.ImagesDocument.Images.Image.Baseimage;
import uk.org.subtrack.imaging.ImagesDocument.Images;
import uk.org.subtrack.imaging.ImagesDocument.Images.Image;
import uk.org.subtrack.imaging.ImagesDocument.Images.Image.Property;
import uk.org.subtrack.imaging.ImagesDocument.Images.Image.Size;

/**
 *
 * @author Mark Elam
 */
public class ImageServlet extends HttpServlet {

    private class ImageInfo {

        public ImageBean imageBean;
        public Image image;

        public ImageInfo(ImageBean iBean, Image i) {
            this.image = i;
            this.imageBean = iBean;
        }
    }
    protected static Log _log = LogFactory.getLog(ImageServlet.class);
    private HashMap<String, ImageInfo> imageMap = new HashMap<String, ImageInfo>();
    private HashMap<String, String> mimeTypes = new HashMap<String, String>();
    private Map<String, BufferedImage> bufferedImageMap = Collections.synchronizedMap(new HashMap<String, BufferedImage>());
    private static String defaultFormat = "png";
    private static String defaultMIMEType = "image/png";
    private static String defaultPropertiesFile = "/imaging.properties";
    private static String defaultImagesFile = "/images.xml";
    private static String IMAGES_FILE = "IMAGES_FILE";
    private static String PROPERTIES_FILE = "PROPERTIES_FILE";

    @Override
    public void init() throws ServletException {
        _log.debug("init called in ImageServlet");
        readProperties();
        readImagesXML();
        super.init();
    }

    private void readProperties() {
        String propertiesFile = getServletConfig().getInitParameter(PROPERTIES_FILE);
        if (propertiesFile == null) {
            propertiesFile = defaultPropertiesFile;
        }
        _log.debug("read properties from " + propertiesFile);
        try {
            Properties props = new Properties();
            props.load(ImageServlet.class.getResourceAsStream(propertiesFile));
            Set keys = props.keySet();
            for (Iterator it = keys.iterator(); it.hasNext();) {
                String key = (String) it.next();
                if (key.startsWith("imaging.format.")) {
                    String format = key.substring(key.lastIndexOf(".") + 1);
                    _log.debug("add to mimeTypes:" + format + ":" + props.get(key));
                    mimeTypes.put(format, (String) props.get(key));
                }
            }
        } catch (IOException ex) {
            _log.debug("no Imaging Properties file found, using defaults", ex);
        }
    }

    private void readImagesXML() {
        String imagesFile = getServletConfig().getInitParameter(IMAGES_FILE);
        if (imagesFile == null) {
            imagesFile = defaultImagesFile;
        }
        _log.debug("read images from " + imagesFile);
        try {
            InputStream is = ImageServlet.class.getResourceAsStream(imagesFile);
            ImagesDocument imagesDocument = ImagesDocument.Factory.parse(is);
            Images images = imagesDocument.getImages();
            Image[] imagesArr = images.getImageArray();
            for (int i = 0; i < imagesArr.length; i++) {
                Image image = imagesArr[i];
                _log.debug(image.getName() + "," + image.getClassName());
                initiliseImage(image);
            }

        } catch (XmlException ex) {
            _log.fatal("Unable to initialise ImagingServlet", ex);
        } catch (IOException ex) {
            _log.fatal("Unable to initialise ImagingServlet", ex);
        }

    }

    @Override
    protected void doGet(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException {
        _log.debug("Path info" + req.getPathInfo());
        ServletOutputStream out = res.getOutputStream();  // binary output!
        String imageName = req.getPathInfo();
        //strip off type
        String format = imageName.substring(imageName.indexOf(".") + 1);
        imageName = imageName.substring(0, imageName.indexOf("."));
        _log.debug("imageName:" + imageName);
        ImageInfo imageInfo = imageMap.get(imageName);
        if (imageInfo != null) {
            Image image = imageInfo.image;
            //copy synchronized which produces local copy for use
            ImageBean imageBean = copy(imageInfo.imageBean);
            BufferedImage bufferedImage = null;
            String fullQueryString = imageName + req.getQueryString();//unique identifier for each image is image name plus parameters
            if ((bufferedImage = retrieveFromImageCache(fullQueryString)) == null) {
                _log.debug("Create new Image for " + fullQueryString);
                //if first request generate Image and store in cache
                setQueryStringParameters(imageBean, req);
                Size size = image.getSize();
                Baseimage baseImage = image.getBaseimage();
                if (baseImage != null) {
                    String path = req.getPathTranslated().substring(0, req.getPathTranslated().lastIndexOf(File.separatorChar));
                    String imageFileName = path + req.getServletPath() + baseImage.getName();
                    bufferedImage = ImageIO.read(new File(imageFileName));
                    imageBean.drawGraphics(bufferedImage.createGraphics(), bufferedImage.getHeight(), bufferedImage.getWidth());
                } else {
                    bufferedImage = new BufferedImage(size.getWidth().intValue(), size.getHeight().intValue(), BufferedImage.TYPE_INT_RGB);
                    imageBean.drawGraphics(bufferedImage.createGraphics(), size.getHeight().intValue(), size.getWidth().intValue());
                }
                storeInImageCache(fullQueryString, bufferedImage);
            }
            String mimeType = mimeTypes.get(format);
            if (mimeType == null) {
                mimeType = defaultMIMEType;
                format = defaultFormat;
            }
            _log.debug("generate image as: ." + format + ":" + mimeType);
            ImageIO.write(bufferedImage, format, out);
            res.setContentType(mimeType);
        }
        // Set to expire far in the past.
        res.setHeader("Expires", "Sat, 6 May 1995 12:00:00 GMT");

        // Set standard HTTP/1.1 no-cache headers.
        res.setHeader("Cache-Control", "no-store, no-cache, must-revalidate");

        // Set IE extended HTTP/1.1 no-cache headers (use addHeader).
        res.addHeader("Cache-Control", "post-check=0, pre-check=0");

        // Set standard HTTP/1.0 no-cache header.
        res.setHeader("Pragma", "no-cache");
    }

    private void initiliseImage(Image image) {
        try {

            Class imageClass = Class.forName(image.getClassName());
            Object imageObj = imageClass.newInstance();
            if (imageObj instanceof ImageBean) {
                ImageBean imageBean = (ImageBean) imageObj;
                //add any properties to new Bean if possible
                Property[] propertys = image.getPropertyArray();
                for (int i = 0; i < propertys.length; i++) {
                    Property property = propertys[i];
                    initialiseProperty(imageBean, property);

                }
                //store in Map
                imageMap.put(image.getName(), new ImageInfo(imageBean, image));
            }


        } catch (InstantiationException ex) {
            _log.debug(null, ex);
        } catch (IllegalAccessException ex) {
            _log.debug(null, ex);
        } catch (ClassNotFoundException ex) {
            _log.debug("Class " + image.getClassName() + " Not found", ex);
        }
    }

    private void initialiseProperty(ImageBean imageBean, Property property) {
        initialiseProperty(imageBean, property.getName(), property.getValue());
    }

    private void initialiseProperty(ImageBean imageBean, String propertyName, String propertyValue) {
        //try field first...
        boolean set = false;
        try {
            Field field = imageBean.getClass().getField(propertyName);
            field.set(imageBean, propertyValue);
            set = true;
        } catch (IllegalArgumentException ex) {
            _log.trace("No field " + propertyName + " found on [" + imageBean.getClass().getName() + "]");
        } catch (IllegalAccessException ex) {
            _log.trace("No field " + propertyName + " found on [" + imageBean.getClass().getName() + "]");
        } catch (NoSuchFieldException ex) {
            _log.trace("No field " + propertyName + " found on [" + imageBean.getClass().getName() + "]");
        } catch (SecurityException ex) {
            _log.trace("No field " + propertyName + " found on [" + imageBean.getClass().getName() + "]");
        }
        //try method otherwise
        if (!set) {
            String methodName = "set" + Character.toUpperCase(propertyName.charAt(0)) + propertyName.substring(1);
            try {
                Method method = imageBean.getClass().getMethod(methodName, new Class[]{String.class});
                method.invoke(imageBean, propertyValue);
            } catch (IllegalAccessException ex) {
                _log.trace("No Method " + methodName + " available on [" + imageBean.getClass().getName() + "]");
            } catch (IllegalArgumentException ex) {
                _log.trace("No Method " + methodName + " available on [" + imageBean.getClass().getName() + "]");
            } catch (InvocationTargetException ex) {
                _log.trace("No Method " + methodName + " available on [" + imageBean.getClass().getName() + "]");
            } catch (NoSuchMethodException ex) {
                _log.trace("No Method " + methodName + " available on [" + imageBean.getClass().getName() + "]");
            } catch (SecurityException ex) {
                _log.trace("No Method " + methodName + " available on [" + imageBean.getClass().getName() + "]");
            }
        }
    }

    private void setQueryStringParameters(ImageBean imageBean, HttpServletRequest req) {
        Map parameters = req.getParameterMap();
        Set properties = parameters.keySet();
        for (Object objProp : properties) {
            String property = (String) objProp;
            initialiseProperty(imageBean, property, ((String[]) parameters.get(property))[0]);
        }

    }

    private BufferedImage retrieveFromImageCache(String fullQueryString) {
        _log.debug("Retrieve previous image for " + fullQueryString);
        return bufferedImageMap.get(fullQueryString);
    }

    private void storeInImageCache(String qString, BufferedImage bufferedImage) {
        _log.debug("Store image for " + qString);
        bufferedImageMap.put(qString, bufferedImage);
    }

    private synchronized ImageBean copy(ImageBean orig) {
        ImageBean imageBean = null;
        try {
            // Write the object out to a byte array
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            ObjectOutputStream out = new ObjectOutputStream(bos);
            out.writeObject(orig);
            out.flush();
            out.close();

            // Make an input stream from the byte array and read
            // a copy of the object back in.
            ObjectInputStream in = new ObjectInputStream(
                    new ByteArrayInputStream(bos.toByteArray()));
            imageBean = (ImageBean) in.readObject();
        } catch (IOException e) {
            _log.fatal("Cannot copy ImageBean", e);
        } catch (ClassNotFoundException e) {
            _log.fatal("Cannot copy ImageBean", e);
        }
        return imageBean;
    }
}
